@@ -3,6 +3,8 @@ import logging
 import time
 import asyncio
 import websockets
+import json
+import urllib
 from dataclasses import asdict
 from aiohttp import web
 from rpc_gateway import messages, errors
@@ -33,15 +35,17 @@ class Gateway:
     SERVER_MESSAGES = (messages.Method.GET, messages.Method.SET, messages.Method.CALL, messages.Method.LOCK,
                        messages.Method.UNLOCK, messages.Method.METADATA)
 
-    def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887):
+    def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887, auth_key: str = 'DEFAULT_KEY'):
         self.host = host
         self.port = port
         self.http_port = http_port
+        self.auth_key = auth_key
         self.logger = logger.getChild(self.__class__.__name__)
         self.websocket_connections: List[WebsocketConnection] = []
         self.websocket: Optional[websockets.WebSocketServer] = None
         self.event_loop = asyncio.get_event_loop()
         self.instances: Dict[str, WebsocketConnection] = {}
+        self.instance_groups: Dict[str, List[str]] = {}
         self.websocket_instances: Dict[WebsocketConnection, List[str]] = {}
 
         self.http_server: Optional[web.Server] = None
@@ -87,6 +91,12 @@ class Gateway:
 
     async def on_connection(self, connection: websockets.WebSocketServerProtocol, path: str):
         self.logger.info(f'New connection from {connection.remote_address} path: {path}')
+        connection_params = urllib.parse.parse_qs(path[1:]) if path != '' else {}
+
+        if 'auth_key' not in connection_params or connection_params['auth_key'][0] != self.auth_key:
+            await connection.close(1008, 'Invalid auth_key')  # 1008 is the "Policy Violation" websocket status code
+            return
+
         websocket_connection = WebsocketConnection(connection, request_handler=self.handle_request, close_handler=self.handle_close)
         self.websocket_connections.append(websocket_connection)
         self.websocket_instances[websocket_connection] = []
@@ -129,11 +139,29 @@ class Gateway:
         server = self.instances[request.data['instance']]
         return await server.request(request)
 
+    async def handle_list_request(self, request: messages.Request) -> messages.Response:
+        if 'group' not in request.data or request.data['group'] is None:
+            instances = self.instance_groups
+        else:
+            group = request.data['group']
+
+            if group not in self.instance_groups:
+                return messages.Response(status=messages.Status.ERROR, data=errors.GroupNotFoundError(f'Group not found: {group}'))
+
+            instances = {group: self.instance_groups[group]}
+
+        return messages.Response(data=instances)
+
     async def handle_register_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
         self.logger.info(f'Registering instances: {request.data}')
-        self.websocket_instances[websocket_connection] += request.data
-        for instance_name in request.data:
+
+        for instance_name, instance_group in request.data:
+            if instance_group not in self.instance_groups:
+                self.instance_groups[instance_group] = []
+
+            self.instance_groups[instance_group].append(instance_name)
             self.instances[instance_name] = websocket_connection
+            self.websocket_instances[websocket_connection].append(instance_name)
 
         return messages.Response()
 
@@ -148,18 +176,29 @@ class Gateway:
         if request.method == messages.Method.REGISTER:
             return await self.handle_register_request(websocket_connection, request)
 
+        if request.method == messages.Method.LIST:
+            return await self.handle_list_request(request)
+
         return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))
 
     async def handle_close(self, websocket_connection: WebsocketConnection):
+        # check if this is a server connection
         if websocket_connection in self.websocket_instances:
             instances = self.websocket_instances[websocket_connection]
 
-            # unlock registered instances
-            await asyncio.gather(*[self._unlock_instance(instance_name) for instance_name in instances])
-
             # remove registered instances
             for instance_name in instances:
                 self.instances.pop(instance_name)
+                groups_to_remove = []
+
+                for group, group_instances in self.instance_groups.items():
+                    group_instances.remove(instance_name)
+
+                    if len(group_instances) == 0:
+                        groups_to_remove.append(group)
+
+                for group in groups_to_remove:
+                    self.instance_groups.pop(group)
 
             self.websocket_instances.pop(websocket_connection)
 
@@ -167,24 +206,45 @@ class Gateway:
         for request_id, response_queue in websocket_connection.receive_queues.items():
             await response_queue.put(messages.Response(status=messages.Status.ERROR, data=errors.ServerConnectionLostError(f'Server connection lost')))
 
+    def check_authentication(self, request: web.BaseRequest) -> bool:
+        if 'Authentication' in request.headers:
+            auth_header = request.headers['Authentication']
+
+            if auth_header.startswith('AuthKey '):
+                auth_type, auth_key = auth_header.split(' ', maxsplit=2)
+
+                return auth_key == self.auth_key
+
+        elif 'auth_key' in request.query:
+            return request.query['auth_key'] == self.auth_key
+
+        return False
+
     async def handle_http_request(self, request: web.BaseRequest) -> web.StreamResponse:
         if request.method != 'POST':
             return web.HTTPMethodNotAllowed(request.method, ['POST'])
 
-        message = await request.json()
+        if not self.check_authentication(request):
+            return web.json_response({'status': 1, 'data': 'Authentication required'}, status=401)
+
+        message_raw = await request.read()
 
         try:
+            self.logger.info(f'Handling request: {message_raw}')
+            message = json.loads(message_raw)
             request = messages.Request(id=next_message_id(), method=message['method'], data=message['data'])
 
             if request.method in self.SERVER_MESSAGES:
                 response = await self.handle_forward_request(request)
             elif request.method == messages.Method.AVAILABLE:
                 response = await self.handle_available_request(request)
+            elif request.method == messages.Method.LIST:
+                response = await self.handle_list_request(request)
             else:
                 response = messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))
 
-        except KeyError:
-            response = messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Invalid message: {message}'))
+        except (KeyError, json.decoder.JSONDecodeError):
+            response = messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Invalid message: {message_raw}'))
 
         if response.status == messages.Status.ERROR:
             response.data = repr(response.data)
@@ -197,9 +257,10 @@ class Gateway:
 
 
 class GatewayClient:
-    def __init__(self, gateway_url: str = 'ws://localhost:8888'):
+    def __init__(self, gateway_url: str = 'ws://localhost:8888', auth_key: str = 'DEFAULT_KEY'):
         self.logger = logger.getChild(self.__class__.__name__)
-        self.gateway_url = gateway_url
+        self.gateway_url = gateway_url[:-1] if gateway_url.endswith('/') else gateway_url
+        self.gateway_auth_key = auth_key
         self.websocket_connection = WebsocketConnection(request_handler=self._log_and_handle_request, close_handler=self._handle_close)
         self.event_loop = asyncio.get_event_loop()
         self.connect_retry_timeout = 2.0
@@ -212,7 +273,7 @@ class GatewayClient:
         while True:
             try:
                 self.logger.info(f'Connecting to {self.gateway_url}')
-                self.connection = await websockets.connect(self.gateway_url)
+                self.connection = await websockets.connect(self.gateway_url + '/socket&auth_key=' + self.gateway_auth_key)
                 return
             except OSError:
                 self.logger.warning(f'Error connecting to {self.gateway_url}, retrying in {self.connect_retry_timeout} seconds')
